"""Utility functions for working with card data."""

import functools
import re
from typing import Tuple

STRICT_BASICS = frozenset({"Plains", "Island", "Swamp", "Mountain", "Forest"})

COLLECTOR_NUMBER_RE = re.compile(r"(?P<prefix>.*-|\D+)?(?P<number>\d+)?(?P<suffix>\D.*)?")


@functools.lru_cache(maxsize=None)
def collector_number_parts(collector_number: str) -> Tuple[str, int, str]:
    """Split a collector number into its parts."""
    match = COLLECTOR_NUMBER_RE.match(collector_number)
    if not match:
        return ("", 0, "")
    prefix: str = match.group("prefix") or ""
    number: int = int(match.group("number")) if match.group("number") else 0
    suffix: str = match.group("suffix") or ""
    return (prefix, number, suffix)


def is_strict_basic(card_name: str) -> bool:
    """Is the card on of the five basic lands (not Snow or Wastes)."""
    return card_name in STRICT_BASICS



"""CSV serializer."""

import csv
import datetime as dt
from pathlib import Path
from typing import Any, ClassVar, Dict, Iterable, Mapping

from mtg_ssm.containers import counts
from mtg_ssm.containers.collection import MagicCollection
from mtg_ssm.containers.counts import CountType
from mtg_ssm.containers.indexes import Oracle
from mtg_ssm.scryfall.models import ScryCard
from mtg_ssm.serialization import interface

CSV_HEADER = ["set", "name", "collector_number", "scryfall_id"] + [ct.value for ct in CountType]


def row_for_card(card: ScryCard, card_count: Mapping[CountType, int]) -> Dict[str, Any]:
    """Given a CardPrinting and counts, return a csv row."""
    return {
        "set": card.set.upper(),
        "name": card.name,
        "collector_number": card.collector_number,
        "scryfall_id": card.id,
        **{ct.value: cnt for ct, cnt in card_count.items() if cnt},
    }


def rows_for_cards(collection: MagicCollection, verbose: bool) -> Iterable[Dict[str, Any]]:
    """Yield csv rows from a collection."""
    for card_set in sorted(
        collection.oracle.index.setcode_to_set.values(),
        key=lambda cset: (cset.released_at or dt.date.min, cset.code),
    ):
        for card in collection.oracle.index.setcode_to_cards[card_set.code]:
            card_count = collection.counts.get(card.id, {})
            if verbose or any(card_count.values()):
                yield row_for_card(card, card_count)


class CsvFullDialect(interface.SerializationDialect):
    """csv collection writing a row for every printing."""

    extension: ClassVar[str] = "csv"
    dialect: ClassVar[str] = "csv"

    verbose: ClassVar[bool] = True

    def write(self, path: Path, collection: MagicCollection) -> None:
        """Write collection to a file."""
        with path.open("wt", encoding="utf-8") as csv_file:
            writer = csv.DictWriter(csv_file, CSV_HEADER)
            writer.writeheader()
            for row in rows_for_cards(collection, self.verbose):
                writer.writerow(row)

    def read(self, path: Path, oracle: Oracle) -> MagicCollection:
        """Read collection from file."""
        with path.open("rt", encoding="utf-8") as csv_file:
            reader = csv.DictReader(csv_file)
            card_counts = counts.aggregate_card_counts(reader, oracle)
        return MagicCollection(oracle=oracle, counts=card_counts)


class CsvTerseDialect(CsvFullDialect):
    """csv collection writing only rows that have counts."""

    dialect: ClassVar[str] = "terse"

    verbose: ClassVar[bool] = False



"""Tests for mtg_ssm.serialization.interface.py."""

from typing import Dict
from unittest import mock

import pytest

from mtg_ssm.serialization import interface


# Subclass registration tests
def test_all_dialects() -> None:
    all_formats = interface.SerializationDialect.dialects()
    assert sorted(all_formats) == [
        ("csv", "csv", mock.ANY),
        ("csv", "terse", mock.ANY),
        ("xlsx", "xlsx", mock.ANY),
    ]


@pytest.mark.parametrize(
    ("extension", "dialect_mapping", "dialect_name"),
    [
        pytest.param("csv", {}, "CsvFullDialect"),
        pytest.param("csv", {"csv": "terse"}, "CsvTerseDialect"),
        pytest.param("csv", {"xlsx": "csv"}, "CsvFullDialect"),
        pytest.param("xlsx", {}, "XlsxDialect"),
        pytest.param(
            "invalid",
            {},
            mock.ANY,
            marks=pytest.mark.xfail(raises=interface.UnknownDialectError),
        ),
        pytest.param(
            "",
            {},
            mock.ANY,
            marks=pytest.mark.xfail(raises=interface.UnknownDialectError),
        ),
    ],
)
def test_extension_lookup(
    extension: str, dialect_mapping: Dict[str, str], dialect_name: str
) -> None:
    serialization_class = interface.SerializationDialect.by_extension(extension, dialect_mapping)
    assert isinstance(serialization_class, type)
    assert issubclass(serialization_class, interface.SerializationDialect)
    assert serialization_class.__name__ == dialect_name



"""Scryfall data fetcher."""

import gzip
import os
from typing import List

import appdirs
import msgspec
from requests_cache import CachedSession, SerializerPipeline, Stage, pickle_serializer

from mtg_ssm.containers.bundles import ScryfallDataSet
from mtg_ssm.scryfall.models import (
    ScryBulkData,
    ScryCard,
    ScryList,
    ScryMigration,
    ScrySet,
)

APP_AUTHOR = "gwax"
APP_NAME = "mtg_ssm"
CACHE_DIR = appdirs.user_cache_dir(APP_NAME, APP_AUTHOR)
CACHE_SERIALIZER = SerializerPipeline(
    [
        pickle_serializer,
        Stage(dumps=gzip.compress, loads=gzip.decompress),
    ],
    is_binary=True,
)
SESSION = CachedSession(
    os.path.join(CACHE_DIR, "requests_cache.sqlite"),  # noqa: PTH118
    backend="sqlite",
    serializer=CACHE_SERIALIZER,
    cache_control=True,
    expire_after=86400,
)

BULK_DATA_ENDPOINT = "https://api.scryfall.com/bulk-data"
SETS_ENDPOINT = "https://api.scryfall.com/sets"
MIGRATIONS_ENDPOINT = "https://api.scryfall.com/migrations"
BULK_TYPE = "default_cards"

REQUESTS_TIMEOUT_SECONDS = 30


def _fetch_endpoint(endpoint: str) -> bytes:
    response = SESSION.get(endpoint)
    response.raise_for_status()
    cached_response = getattr(response, "from_cache", False)
    print(f'Fetched {endpoint}{" [CACHED]" if cached_response else ""}')
    return response.content


def scryfetch() -> ScryfallDataSet:
    """Retrieve and deserialize Scryfall object data."""
    print("Reading data from scryfall")
    bulk_data = msgspec.json.decode(
        _fetch_endpoint(BULK_DATA_ENDPOINT), type=ScryList[ScryBulkData]
    ).data

    scrylistset_decoder = msgspec.json.Decoder(ScryList[ScrySet])
    sets_list = scrylistset_decoder.decode(_fetch_endpoint(SETS_ENDPOINT))
    sets_data = sets_list.data
    while sets_list.has_more and sets_list.next_page is not None:
        sets_list = scrylistset_decoder.decode(_fetch_endpoint(sets_list.next_page))
        sets_data += sets_list.data

    scrylistmigration_decoder = msgspec.json.Decoder(ScryList[ScryMigration])
    migrations_list = scrylistmigration_decoder.decode(_fetch_endpoint(MIGRATIONS_ENDPOINT))
    migrations_data = migrations_list.data
    while migrations_list.has_more and migrations_list.next_page is not None:
        migrations_list = scrylistmigration_decoder.decode(
            _fetch_endpoint(migrations_list.next_page)
        )
        migrations_data += migrations_list.data

    [cards_endpoint] = [bd.download_uri for bd in bulk_data if bd.type == BULK_TYPE]
    cards_data = msgspec.json.decode(_fetch_endpoint(cards_endpoint), type=List[ScryCard])

    return ScryfallDataSet(sets=sets_data, cards=cards_data, migrations=migrations_data)



"""Tests for mtg_ssm.mtg.util."""

from typing import Dict, Tuple
from uuid import UUID

import pytest

from mtg_ssm.mtg import util
from mtg_ssm.scryfall.models import ScryCard


@pytest.mark.parametrize(
    ("name", "expected"),
    [
        pytest.param("Forest", True),
        pytest.param("Snow-Covered Forest", False),
        pytest.param("Wastes", False),
        pytest.param("Abattoir Ghoul", False),
    ],
)
def test_is_strict_basic(name: str, expected: bool) -> None:
    assert util.is_strict_basic(name) is expected


@pytest.mark.parametrize(
    ("name", "card_id", "number_parts"),
    [
        pytest.param("Thallid", UUID("4caaf31b-86a9-485b-8da7-d5b526ed1233"), ("", 74, "a")),
        pytest.param("Dark Ritual", UUID("ebb6664d-23ca-456e-9916-afcd6f26aa7f"), ("", 98, "")),
        pytest.param(
            "Stairs to Infinity", UUID("57f25ead-b3ec-4c40-972d-d750ed2f5319"), ("P", 1, "")
        ),
        pytest.param(
            "Ertai, the Corrupted", UUID("66b950d9-8fef-4deb-b51b-26edb90abc56"), ("", 107, "")
        ),
        pytest.param(
            "Ertai, the Corrupted", UUID("fbbfeb32-1654-4bf6-9a38-891f1a03e02b"), ("", 107, "★")
        ),
        pytest.param("Viscera Seer", UUID("dba1cf83-e13d-401e-b76f-b12a51b307f9"), ("VS", 0, "")),
        pytest.param(
            "Anointed Procession", UUID("c9d6af7b-41a5-4077-81ed-7560548dcf3e"), ("AKH-", 2, "")
        ),
        pytest.param("Terror", UUID("f8c2f285-141b-479b-aca6-ceacf60e3f6a"), ("10E-", 182, "")),
    ],
)
def test_collector_number_parts(
    id_to_card: Dict[UUID, ScryCard],
    name: str,
    card_id: UUID,
    number_parts: Tuple[str, int, str],
) -> None:
    card = id_to_card[card_id]
    assert card.name == name
    assert number_parts == util.collector_number_parts(card.collector_number)



"""Tests for mtg_ssm.containers.indexes."""

from uuid import UUID

from mtg_ssm.containers.bundles import ScryfallDataSet
from mtg_ssm.containers.indexes import ScryfallDataIndex


def test_load_data(scryfall_data: ScryfallDataSet) -> None:
    index = ScryfallDataIndex()
    index.load_data(scryfall_data)


def test_id_to_card(scryfall_data: ScryfallDataSet) -> None:
    index = ScryfallDataIndex()
    index.load_data(scryfall_data)
    assert index.id_to_card[UUID("59cf0906-04fa-4b30-a7a6-3d117931154f")].name == "Abattoir Ghoul"


def test_name_to_cards(scryfall_data: ScryfallDataSet) -> None:
    index = ScryfallDataIndex()
    index.load_data(scryfall_data)

    # set sorting
    assert [c.set for c in index.name_to_cards["Dark Ritual"]] == ["hop", "ice", "lea"]

    # collector number sorting
    assert [(c.set, c.collector_number) for c in index.name_to_cards["Forest"]] == [
        ("ice", "380"),
        ("ice", "381"),
        ("ice", "382"),
        ("isd", "262"),
        ("isd", "263"),
        ("isd", "264"),
        ("lea", "294"),
        ("lea", "295"),
    ]

    # collector number variant sorting
    assert [(c.set, c.collector_number) for c in index.name_to_cards["Thallid"]] == [
        ("fem", "74a"),
        ("fem", "74b"),
        ("fem", "74c"),
        ("fem", "74d"),
        ("mma", "167"),
    ]


def test_setcode_to_cards(scryfall_data: ScryfallDataSet) -> None:
    index = ScryfallDataIndex()
    index.load_data(scryfall_data)

    # collector number sorting
    assert [(c.name, c.collector_number) for c in index.setcode_to_cards["isd"]] == [
        ("Delver of Secrets // Insectile Aberration", "51"),
        ("Abattoir Ghoul", "85"),
        ("Forest", "262"),
        ("Forest", "263"),
        ("Forest", "264"),
    ]

    # collector number variant sorting
    assert [c.collector_number for c in index.setcode_to_cards["fem"]] == [
        "74a",
        "74b",
        "74c",
        "74d",
    ]

    # The List Sorting
    assert [(c.collector_number, c.name) for c in index.setcode_to_cards["plst"]] == [
        ("10E-20", "High Ground"),
        ("10E-182", "Terror"),
        ("AKH-2", "Anointed Procession"),
        ("AKH-3", "Anointer Priest"),
        ("ARB-1", "Ardent Plea"),
    ]


def test_id_to_setindex(scryfall_data: ScryfallDataSet) -> None:
    index = ScryfallDataIndex()
    index.load_data(scryfall_data)

    assert {c.id: index.id_to_setindex[c.id] for c in index.setcode_to_cards["isd"]} == {
        UUID("11bf83bb-c95b-4b4f-9a56-ce7a1816307a"): 0,
        UUID("59cf0906-04fa-4b30-a7a6-3d117931154f"): 1,
        UUID("b606f644-1728-4cb3-90ed-121838875de1"): 2,
        UUID("16f52885-1f01-4f06-90a8-1a0ecf291ab5"): 3,
        UUID("4dea3762-c6ae-4304-aee4-6c3f56685319"): 4,
    }


def test_setcode_to_set(scryfall_data: ScryfallDataSet) -> None:
    index = ScryfallDataIndex()
    index.load_data(scryfall_data)

    assert index.setcode_to_set["isd"].name == "Innistrad"


def test_snnma_to_id(scryfall_data: ScryfallDataSet) -> None:
    index = ScryfallDataIndex()
    index.load_data(scryfall_data)

    assert {
        k
        for k, v in index.snnma_to_id.items()
        if v == {UUID("59cf0906-04fa-4b30-a7a6-3d117931154f")}
    } == {
        ("isd", "Abattoir Ghoul", "85", 222911, None),
        ("isd", "Abattoir Ghoul", "85", None, None),
        ("isd", "Abattoir Ghoul", None, 222911, None),
        ("isd", "Abattoir Ghoul", None, None, None),
        ("isd", "Abattoir Ghoul", "85", 222911, "Volkan Baǵa"),
        ("isd", "Abattoir Ghoul", "85", None, "Volkan Baǵa"),
        ("isd", "Abattoir Ghoul", None, 222911, "Volkan Baǵa"),
        ("isd", "Abattoir Ghoul", None, None, "Volkan Baǵa"),
        (None, "Abattoir Ghoul", "85", 222911, None),
        (None, "Abattoir Ghoul", "85", None, None),
        (None, "Abattoir Ghoul", None, 222911, None),
        (None, "Abattoir Ghoul", None, None, None),
        (None, "Abattoir Ghoul", "85", 222911, "Volkan Baǵa"),
        (None, "Abattoir Ghoul", "85", None, "Volkan Baǵa"),
        (None, "Abattoir Ghoul", None, 222911, "Volkan Baǵa"),
        (None, "Abattoir Ghoul", None, None, "Volkan Baǵa"),
    }

    assert {
        k
        for k, v in index.snnma_to_id.items()
        if v == {UUID("11bf83bb-c95b-4b4f-9a56-ce7a1816307a")}
    } == {
        ("isd", "Delver of Secrets // Insectile Aberration", "51", 226749, None),
        ("isd", "Delver of Secrets // Insectile Aberration", "51", 226755, None),
        ("isd", "Delver of Secrets // Insectile Aberration", "51", None, None),
        ("isd", "Delver of Secrets // Insectile Aberration", None, 226749, None),
        ("isd", "Delver of Secrets // Insectile Aberration", None, 226755, None),
        ("isd", "Delver of Secrets // Insectile Aberration", None, None, None),
        ("isd", "Delver of Secrets", "51", 226749, None),
        ("isd", "Delver of Secrets", "51", 226755, None),
        ("isd", "Delver of Secrets", "51", None, None),
        ("isd", "Delver of Secrets", "51a", 226749, None),
        ("isd", "Delver of Secrets", "51a", 226755, None),
        ("isd", "Delver of Secrets", "51a", None, None),
        ("isd", "Delver of Secrets", None, 226749, None),
        ("isd", "Delver of Secrets", None, 226755, None),
        ("isd", "Delver of Secrets", None, None, None),
        ("isd", "Insectile Aberration", "51", 226749, None),
        ("isd", "Insectile Aberration", "51", 226755, None),
        ("isd", "Insectile Aberration", "51", None, None),
        ("isd", "Insectile Aberration", "51b", 226749, None),
        ("isd", "Insectile Aberration", "51b", 226755, None),
        ("isd", "Insectile Aberration", "51b", None, None),
        ("isd", "Insectile Aberration", None, 226749, None),
        ("isd", "Insectile Aberration", None, 226755, None),
        ("isd", "Insectile Aberration", None, None, None),
        ("isd", "Delver of Secrets // Insectile Aberration", "51", 226749, "Nils Hamm"),
        ("isd", "Delver of Secrets // Insectile Aberration", "51", 226755, "Nils Hamm"),
        ("isd", "Delver of Secrets // Insectile Aberration", "51", None, "Nils Hamm"),
        ("isd", "Delver of Secrets // Insectile Aberration", None, 226749, "Nils Hamm"),
        ("isd", "Delver of Secrets // Insectile Aberration", None, 226755, "Nils Hamm"),
        ("isd", "Delver of Secrets // Insectile Aberration", None, None, "Nils Hamm"),
        ("isd", "Delver of Secrets", "51", 226749, "Nils Hamm"),
        ("isd", "Delver of Secrets", "51", 226755, "Nils Hamm"),
        ("isd", "Delver of Secrets", "51", None, "Nils Hamm"),
        ("isd", "Delver of Secrets", "51a", 226749, "Nils Hamm"),
        ("isd", "Delver of Secrets", "51a", 226755, "Nils Hamm"),
        ("isd", "Delver of Secrets", "51a", None, "Nils Hamm"),
        ("isd", "Delver of Secrets", None, 226749, "Nils Hamm"),
        ("isd", "Delver of Secrets", None, 226755, "Nils Hamm"),
        ("isd", "Delver of Secrets", None, None, "Nils Hamm"),
        ("isd", "Insectile Aberration", "51", 226749, "Nils Hamm"),
        ("isd", "Insectile Aberration", "51", 226755, "Nils Hamm"),
        ("isd", "Insectile Aberration", "51", None, "Nils Hamm"),
        ("isd", "Insectile Aberration", "51b", 226749, "Nils Hamm"),
        ("isd", "Insectile Aberration", "51b", 226755, "Nils Hamm"),
        ("isd", "Insectile Aberration", "51b", None, "Nils Hamm"),
        ("isd", "Insectile Aberration", None, 226749, "Nils Hamm"),
        ("isd", "Insectile Aberration", None, 226755, "Nils Hamm"),
        ("isd", "Insectile Aberration", None, None, "Nils Hamm"),
        (None, "Delver of Secrets // Insectile Aberration", "51", 226749, None),
        (None, "Delver of Secrets // Insectile Aberration", "51", 226755, None),
        (None, "Delver of Secrets // Insectile Aberration", "51", None, None),
        (None, "Delver of Secrets // Insectile Aberration", None, 226749, None),
        (None, "Delver of Secrets // Insectile Aberration", None, 226755, None),
        (None, "Delver of Secrets // Insectile Aberration", None, None, None),
        (None, "Delver of Secrets", "51", 226749, None),
        (None, "Delver of Secrets", "51", 226755, None),
        (None, "Delver of Secrets", "51", None, None),
        (None, "Delver of Secrets", "51a", 226749, None),
        (None, "Delver of Secrets", "51a", 226755, None),
        (None, "Delver of Secrets", "51a", None, None),
        (None, "Delver of Secrets", None, 226749, None),
        (None, "Delver of Secrets", None, 226755, None),
        (None, "Delver of Secrets", None, None, None),
        (None, "Insectile Aberration", "51", 226749, None),
        (None, "Insectile Aberration", "51", 226755, None),
        (None, "Insectile Aberration", "51", None, None),
        (None, "Insectile Aberration", "51b", 226749, None),
        (None, "Insectile Aberration", "51b", 226755, None),
        (None, "Insectile Aberration", "51b", None, None),
        (None, "Insectile Aberration", None, 226749, None),
        (None, "Insectile Aberration", None, 226755, None),
        (None, "Insectile Aberration", None, None, None),
        (None, "Delver of Secrets // Insectile Aberration", "51", 226749, "Nils Hamm"),
        (None, "Delver of Secrets // Insectile Aberration", "51", 226755, "Nils Hamm"),
        (None, "Delver of Secrets // Insectile Aberration", "51", None, "Nils Hamm"),
        (None, "Delver of Secrets // Insectile Aberration", None, 226749, "Nils Hamm"),
        (None, "Delver of Secrets // Insectile Aberration", None, 226755, "Nils Hamm"),
        (None, "Delver of Secrets // Insectile Aberration", None, None, "Nils Hamm"),
        (None, "Delver of Secrets", "51", 226749, "Nils Hamm"),
        (None, "Delver of Secrets", "51", 226755, "Nils Hamm"),
        (None, "Delver of Secrets", "51", None, "Nils Hamm"),
        (None, "Delver of Secrets", "51a", 226749, "Nils Hamm"),
        (None, "Delver of Secrets", "51a", 226755, "Nils Hamm"),
        (None, "Delver of Secrets", "51a", None, "Nils Hamm"),
        (None, "Delver of Secrets", None, 226749, "Nils Hamm"),
        (None, "Delver of Secrets", None, 226755, "Nils Hamm"),
        (None, "Delver of Secrets", None, None, "Nils Hamm"),
        (None, "Insectile Aberration", "51", 226749, "Nils Hamm"),
        (None, "Insectile Aberration", "51", 226755, "Nils Hamm"),
        (None, "Insectile Aberration", "51", None, "Nils Hamm"),
        (None, "Insectile Aberration", "51b", 226749, "Nils Hamm"),
        (None, "Insectile Aberration", "51b", 226755, "Nils Hamm"),
        (None, "Insectile Aberration", "51b", None, "Nils Hamm"),
        (None, "Insectile Aberration", None, 226749, "Nils Hamm"),
        (None, "Insectile Aberration", None, 226755, "Nils Hamm"),
        (None, "Insectile Aberration", None, None, "Nils Hamm"),
    }

    assert {
        k
        for k, v in index.snnma_to_id.items()
        if UUID("4caaf31b-86a9-485b-8da7-d5b526ed1233") in v
    } == {
        ("fem", "Thallid", "74a", 1924, None),
        ("fem", "Thallid", "74a", None, None),
        ("fem", "Thallid", None, 1924, None),
        ("fem", "Thallid", None, None, None),
        ("fem", "Thallid", "74a", 1924, "Edward P. Beard, Jr."),
        ("fem", "Thallid", "74a", None, "Edward P. Beard, Jr."),
        ("fem", "Thallid", None, 1924, "Edward P. Beard, Jr."),
        ("fem", "Thallid", None, None, "Edward P. Beard, Jr."),
        (None, "Thallid", "74a", 1924, None),
        (None, "Thallid", "74a", None, None),
        (None, "Thallid", None, 1924, None),
        (None, "Thallid", None, None, None),
        (None, "Thallid", "74a", 1924, "Edward P. Beard, Jr."),
        (None, "Thallid", "74a", None, "Edward P. Beard, Jr."),
        (None, "Thallid", None, 1924, "Edward P. Beard, Jr."),
        (None, "Thallid", None, None, "Edward P. Beard, Jr."),
    }
    assert index.snnma_to_id[("fem", "Thallid", None, None, None)] == {
        UUID("4caaf31b-86a9-485b-8da7-d5b526ed1233"),
        UUID("80f8f778-ae31-45cd-b27f-f93a07853ede"),
        UUID("2cf2f3da-9101-439d-8caa-910ff40bfbb3"),
        UUID("01827286-b104-41c5-bac9-7c38414bc40e"),
    }



"""Tests for mtg_ssm.containers.legacy."""

from typing import Any, Dict
from uuid import UUID

import pytest

from mtg_ssm.containers import legacy
from mtg_ssm.containers.bundles import ScryfallDataSet
from mtg_ssm.containers.indexes import Oracle


@pytest.fixture(scope="session")
def oracle(scryfall_data: ScryfallDataSet) -> Oracle:
    """Oracle fixture."""
    return Oracle(scryfall_data)


@pytest.mark.parametrize(
    ("card_row", "expected"),
    [
        pytest.param({}, {}),
        pytest.param({"gar": "bage"}, {}),
        pytest.param({"nonfoil": "1", "foil": "3"}, {"nonfoil": 1, "foil": 3}),
        pytest.param({"nonfoil": "0", "foil": "0"}, {}),
        pytest.param({"copies": "1", "foils": "3"}, {"nonfoil": 1, "foil": 3}),
        pytest.param(
            {"copies": "1", "nonfoil": "3", "foils": "5", "foil": "7"},
            {"nonfoil": 4, "foil": 12},
        ),
    ],
)
def test_extract_counts(card_row: Dict[str, Any], expected: Dict[str, int]) -> None:
    assert legacy.extract_counts(card_row) == expected


@pytest.mark.parametrize(
    ("card_row", "expected"),
    [
        pytest.param(
            {},
            None,
            marks=pytest.mark.xfail(raises=legacy.NoMatchError),
            id="not found",
        ),
        pytest.param(
            {"set": "FEM", "name": "Thallid", "number": "74a"},
            UUID("4caaf31b-86a9-485b-8da7-d5b526ed1233"),
            id="snn__",
        ),
        pytest.param(
            {"set": "FEM", "name": "Thallid", "multiverseid": 1924},
            UUID("4caaf31b-86a9-485b-8da7-d5b526ed1233"),
            id="sn_m_",
        ),
        pytest.param(
            {"set": "FEM", "name": "Thallid", "artist": "Edward P. Beard, Jr."},
            UUID("4caaf31b-86a9-485b-8da7-d5b526ed1233"),
            id="sn__a",
        ),
        pytest.param(
            {"set": "MMA", "name": "Thallid"},
            UUID("69d20d28-76e9-4e6e-95c3-f88c51dfabfd"),
            id="sn___",
        ),
        pytest.param(
            {"name": "Thallid", "artist": "Trevor Claxton"},
            UUID("69d20d28-76e9-4e6e-95c3-f88c51dfabfd"),
            id="_n__a",
        ),
        pytest.param(
            {"set": "FEM", "name": "Thallid"},
            None,
            marks=pytest.mark.xfail(raises=legacy.MultipleMatchError),
            id="sn__ multiple",
        ),
        pytest.param(
            {"set": "pMBR", "name": "Black Sun's Zenith", "artist": "James Paick"},
            UUID("dd88131a-2811-4a1f-bb9a-c82e12c1493b"),
            id="set remap",
        ),
        pytest.param(
            {"name": "Dragonscale General", "artist": "William Murai"},
            UUID("6daabdc2-e8a8-41a6-a9f0-1973d9c31d39"),
            id="artist remap",
        ),
    ],
)
def test_find_scryfall_id(card_row: Dict[str, Any], expected: UUID, oracle: Oracle) -> None:
    assert legacy.find_scryfall_id(card_row, oracle) == expected


@pytest.mark.parametrize(
    ("card_row", "expected"),
    [
        pytest.param({}, {}),
        pytest.param({"set": "MMA", "name": "Thallid"}, {}),
        pytest.param(
            {"set": "MMA", "name": "Thallid", "copies": "1"},
            {"scryfall_id": UUID("69d20d28-76e9-4e6e-95c3-f88c51dfabfd"), "nonfoil": 1},
        ),
    ],
)
def test_coerce_row(card_row: Dict[str, Any], expected: Dict[str, Any], oracle: Oracle) -> None:
    assert legacy.coerce_row(card_row, oracle) == expected
